#include <iostream>
#include <vector>
#include <eigen3/Eigen/Dense>
#include <eigen3/Eigen/Sparse>

using namespace Eigen;
using std::cout;
using std::endl;

class MatrixCOO {
	friend class MatrixCRS; // MatrixCRS will be able to access private members

private:
	std::vector<Triplet<double>> list;

public:
	MatrixCOO() {
	}

	MatrixCOO(const std::vector<Triplet<double>> &list_) {
		// constructor from raw list; performs deep copy
		list = list_;
	}

	MatrixCOO(const MatrixCOO &A) {
		// copy constructor
		MatrixCOO(A.list);
	}

	MatrixCOO(const MatrixXd &A) {
	// constructor from Eigen dense matrix
		for (int i=0; i < A.rows(); i++) {
			for (int j=0; j < A.cols(); j++) {
				if (A(i,j) != 0) {
					Triplet<double> t(i, j, A(i,j));
					list.push_back(t);
				}
			}
		}
	}

	void sort_byRow() {
		// std::sort can sort by applying a specific comparison function,
		// which is what is passed hereafter as third parameter using
		// the lambda expressions syntax.
		std::sort(list.begin(), list.end(),
			[] (auto t1, auto t2) {
				return t1.row() < t2.row();
			});
	}

	int cols() { // returns maximum column which appears among the triplets
		int maxCol = -1;
		for (auto t : list) {
			if (maxCol < t.col()) {
				maxCol = t.col();
			}
		}
		return maxCol+1;
	}

	int rows() {
	// returns maximum row which appears among the triplets
		int maxRow = -1;
		for (auto t : list) {
			if (maxRow < t.row()) {
				maxRow = t.row();
			}
		}
		return maxRow+1;
	}

	MatrixXd toDense() {
	// returns an Eigen dense version
		MatrixXd dense = MatrixXd::Zero(rows(), cols());
		for (auto t : list) {
			dense(t.row(), t.col()) += t.value();
		}
		return dense;
	}

	friend std::ostream& operator<<(std::ostream &os, const MatrixCOO& A) {
		// overload operator<< for std output printing
		os << "[ ";
		for (auto t : A.list) {
			os << "(" << t.row() << " " << t.col() << " " << t.value() << ") ";
		}
		os << "]" << endl;
		return os;
	}
};

class MatrixCRS {
	/*
	 *	The conventions used are such that given a dense matrix A,
	 *	for its CRS format it holds that:
	 *	val contains the non-zero elements of A, as read from left to right then
	 *	from top to bottom,
	 *	col_ind contains in position j the column index of the value in position
	 *	j of val,
	 *	row_ptr has in position j the index in val of the first non-zero element 
	 *	of the j-th	row of A, and if the j-th row is made of only zeros it has the
	 *	same element which is in position j+1 of row_ptr; finally, the last element
	 *	of row_ptr is the number of non-zero elements of A.
	*/

private:
	std::vector<double> val;
	std::vector<int> col_ind;
	std::vector<int> row_ptr;

public:
	MatrixCRS() {
		row_ptr.push_back(0);
	}

	MatrixCRS(const MatrixCRS &A) {
		// copy constructor
		val = A.val;
		col_ind = A.col_ind;
		row_ptr = A.row_ptr;
	}

	MatrixCRS(const MatrixXd &A) {
		// constructor from Eigen dense matrix
		int nnz = 0; // number of non-zero elements of A encountered
		int prevNotEmptyRow = -1;

		// fill up val and colind by looping through triplets
		for (int i=0; i < A.rows(); i++) {

			for (int j=0; j < A.cols(); j++) {
				if (A(i,j) != 0) {

					while (prevNotEmptyRow < i) {
						// some additional care must be taken
						// for empty rows when filling rowptr
						row_ptr.push_back(nnz);
						prevNotEmptyRow++;
					}

					nnz++;
					val.push_back(A(i,j));
					col_ind.push_back(j);
				}
			}
		}
		row_ptr.push_back(nnz);
	}


	MatrixCRS(MatrixCOO &A) {
		// constructor from COO matrix
		A.sort_byRow(); // bring A to a more maneagable form

		int prevNotEmptyRow = -1;
		auto &l = A.list; // renaming for ease of use

		for (int i=0; i < l.size(); i++) {

			while (prevNotEmptyRow < l[i].row()) {
				// if there are some empty rows, fill up rowptr
				// following the conventions.
				prevNotEmptyRow++;
				row_ptr.push_back(i);
			}

			double currVal = l[i].value();

			while (i < l.size()-1 && l[i].col() == l[i+1].col() &&
				   l[i].row() == l[i+1].row()) {
				// add up the values corresponding to the same indices
				currVal += l[i].value();
				i++;
			}

			val.push_back(currVal);
			col_ind.push_back(l[i].col());
		}
		row_ptr.push_back(val.size());
	}

	int cols() {
		// returns maximum column which appears 
		int maxCol = -1;
		for (auto c : col_ind) {
			if (maxCol < c) {
				maxCol = c;
			}
		}
		return maxCol+1;
	}

	int rows() {
		// returns maximum row which appears 
		return row_ptr.size()-1;
	}

	MatrixXd toDense() {
		// conversion to Eigen dense matrix
		int nRows = rows();
		MatrixXd dense = MatrixXd::Zero(nRows, cols());

		for (int row = nRows-1; row >= 0; row--) {
			for (int i = row_ptr[row]; i < row_ptr[row+1]; i++) {
				dense(row, col_ind[i]) = val[i];
			}
		}

		return dense;
	}

	friend std::ostream& operator<<(std::ostream &os, const MatrixCRS& A) {

		os << "val     = [ ";
		for (auto x : A.val) {
			os << x << " ";
		}
		os << "]" << endl;

		os << "col_ind = [ ";
		for (auto x : A.col_ind) {
			os << x << " ";
		}
		os << "]" << endl;

		os << "row_ptr = [ ";
		for (auto x : A.row_ptr) {
			os << x << " ";
		}
		os << "]" << endl;

		return os;
	}
};

// extensive correctness tests
int main() {
	MatrixXd A0(1,1);
	A0<< 0;
	MatrixXd A1(1,1);
	A1<< 1;
	MatrixXd A2(2,1);
	A2<< 1,
		 2;
	MatrixXd A3(1,2);
	A3<< 1,2;
	MatrixXd A4(3,3);
	A4<< 1, 0, 0,
		 0, 2, 0,
		 0, 0, 3;
	MatrixXd A5(4,4);
	A5<< 0, 2, 0, 0,
		 0, 0, 0, 0,
		 1, 0, 3, 0,
		 0, 0, 0, 0;
	MatrixXd A6(4,4);
	A6<< 0, 0, 0, 0,
		 0, 0, 0, 0,
		 0, 0, 7, 0,
		 0, 0, 0, 0;
	MatrixXd A7(9,9);
	A7<< 0, 0, 0, 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0, 0, 0, 0;
	MatrixXd A8(9,9);
	A8<< 0, 0, 0, 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 6, 0,.5, 0, 0,
		 0, 0, 0, 0, 0, 0, 0, 0, 0,
		 0, 0, 1, 0, 0, 0,-1, 0, 0,
		 0, 0, 0, 0, 0, 0, 0, 0, 0,
		 0, 9, 8, 0, 0, 0, 3, 0, 0,
		 0, 0, 0, 0, 0, 0, 0, 0, 0,
		 0, 0, 7, 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0, 0, 0, 0;

	MatrixCOO B0(A0);
	MatrixCOO B1(A1);
	MatrixCOO B2(A2);
	MatrixCOO B3(A3);
	MatrixCOO B4(A4);
	MatrixCOO B5(A5);
	MatrixCOO B6(A6);
	MatrixCOO B7(A7);
	MatrixCOO B8(A8);

	MatrixCRS C0(B0);
	MatrixCRS C1(B1);
	MatrixCRS C2(B2);
	MatrixCRS C3(B3);
	MatrixCRS C4(B4);
	MatrixCRS C5(B5);
	MatrixCRS C6(B6);
	MatrixCRS C7(B7);
	MatrixCRS C8(B8);

	MatrixXd D0 = C0.toDense();
	MatrixXd D1 = C1.toDense();
	MatrixXd D2 = C2.toDense();
	MatrixXd D3 = C3.toDense();
	MatrixXd D4 = C4.toDense();
	MatrixXd D5 = C5.toDense();
	MatrixXd D6 = C6.toDense();
	MatrixXd D7 = C7.toDense();
	MatrixXd D8 = C8.toDense();

	cout << "Print format:\nMatrix name\nStarting dense matrix\n->\nCOO format\n->\nCRS format\n->\nback to dense format.\n\n";
	cout << "A0\n" << A0 << endl << "->" << endl << B0 << "->" << endl << C0 << "->" << endl << D0 << endl << endl;
	cout << "A1\n" << A1 << endl << "->" << endl << B1 << "->" << endl << C1 << "->" << endl << D1 << endl << endl;
	cout << "A2\n" << A2 << endl << "->" << endl << B2 << "->" << endl << C2 << "->" << endl << D2 << endl << endl;
	cout << "A3\n" << A3 << endl << "->" << endl << B3 << "->" << endl << C3 << "->" << endl << D3 << endl << endl;
	cout << "A4\n" << A4 << endl << "->" << endl << B4 << "->" << endl << C4 << "->" << endl << D4 << endl << endl;
	cout << "A5\n" << A5 << endl << "->" << endl << B5 << "->" << endl << C5 << "->" << endl << D5 << endl << endl;
	cout << "A6\n" << A6 << endl << "->" << endl << B6 << "->" << endl << C6 << "->" << endl << D6 << endl << endl;
	cout << "A7\n" << A7 << endl << "->" << endl << B7 << "->" << endl << C7 << "->" << endl << D7 << endl << endl;
	cout << "A8\n" << A8 << endl << "->" << endl << B8 << "->" << endl << C8 << "->" << endl << D8 << endl << endl;

	std::vector<Triplet<double>> list;
	list.push_back(Triplet<double>(3,2,-1));
	list.push_back(Triplet<double>(3,2,0.9));
	list.push_back(Triplet<double>(0,0,1));
	list.push_back(Triplet<double>(3,2,0.1));
	list.push_back(Triplet<double>(0,0,1));
	list.push_back(Triplet<double>(1,5,0.1));
	list.push_back(Triplet<double>(0,0,1));
	list.push_back(Triplet<double>(1,5,0.1));
	MatrixCOO X(list);
	cout << "Converting " << X << "to dense we obtain\n" << X.toDense() << endl;
	cout << "If we convert to CRS however,";
	MatrixCRS Y(X);
	cout << " the order of the triplet is rearranged: " << X;

}
